require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/finman/financialtransaction/payment_create'
require 'uu/finman/financialtransaction/payment_attributes'
require 'uu/finman/financialtransaction/payment_set_attributes'

# Module Unicorn Universe
module UU
  # Module Finman
  module Finman
    # Module of finman Financial transaction
    module FinancialTransaction
      #Module of Payment
      module Payment
        extend self

        # Service path
        PATH = 'uu/finman/financialtransaction/FinancialTransaction'

        # Creates a new Payment. The command creates a new Payment for specified business case
        #
        # <b>Required attributes:</b>
        #   control_business_case_uri, counterparty_business_case_uri, name, amount_without_vat, vat_rate or amount_with_vat, financial_transaction_type, is_transaction_credit,
        #   realization_probability, expected_pay_date, due_date, protocol_date, vat_date, transaction_currency, (credit_account_number and credit_side_currrency) or credit_account_uri,
        #   (debit_account_number and debit_side_currency) or debit_account_uri
        # Exactly one of attributes <code>credit_account_number</code> and <code>credit_account_uri</code> has to be set.
        # Exactly one of attributes <code>debit_account_number</code> and <code>debit_account_uri</code> has to be set.
        # Exactly one of attributes <code>vat_rate</code> and <code>amount_with_vat</code> has to be set. Both attributes can be set simultaneously only if following statement is true : <code>amount_without_vat</code> * <code>vat_rate</code> = <code>amount_with_rate</code>.
        # Attribute <code>credit_side_currency</code> and <code>debit_side_currency</code> must be set if respective account is set by number.
        # Attribute <code>transaction_currency</code> must be set and must be one from <code>credit_side_currency</code> and <code>debit_side_currency</code>, with exception if both <code>credit_side_currency</code> and <code>debit_side_currency</code> are equal.
        #
        # @param [UU::OS::UESURI] control_business_case_uri UESURI of the business case of the control side to create Payment for.
        # @param [UU::Finman::FinancialTransaction::PaymentCreate] payment DTO containing attributes of the new Payment
        #
        # @return [UU::OS::UESURI] UESURI of the new Payment. In case of setting <code>dry_run</code> parameter to <code>true</code>, <code>UU::OS::UESURI::NIL_URI</code> is returned instead.
        #
        #
        # @example
        #      # Prepare Business Case URI, which we want to use to create Payment.
        #      control_business_case_uri = UU::OS::UESURI.new('ues:TERRITORY:BUSINESS_CASE');
        #
        #      # Create Payment
        #          payment_uri = UU::Finman::FinancialTransaction::Payment.create(control_business_case_uri,
        #          :name => "Platba ",
        #          :credit_description => "Kreditní popis",
        #          :debit_description => "Debetní popis",
        #          :credit_financial_category_uri => UU::OS::UESURI.new("ues:TERRITORY:APPS.UU.FINMAN/CONFIGURATION:SYS.FIN.CONFIG/TX_TYPE/CR.TXWER"),
        #          :debit_financial_category_uri => UU::OS::UESURI.new("ues:TERRITORY:APPS.UU.FINMAN/CONFIGURATION:SYS.FIN.CONFIG/TX_TYPE/DR.TXOTR"),
        #          :realization_probability => 50,
        #          :is_transaction_credit => true,
        #          :financial_transaction_type => "NONCASH",
        #          :amount_without_vat => 100,
        #          :amount_with_vat =>  120,
        #          :vat_rate => 20,
        #          :is_ready_for_acceptance => true,
        #          :invoice_number => "Číslo faktury ",
        #          :constant_symbol => "Konstantní symbol",
        #          :variable_symbol => "Variabilní symbol",
        #          :specific_symbol => "Specifický symbol",
        #          :for_acceptance_date => "2014-11-02",
        #          :protocol_date => "2014-11-02",
        #          :vat_date => "2014-11-02",
        #          :expected_pay_date => "2014-11-02",
        #          :counterparty_business_case_uri =>  UU::OS::UESURI.new("ues:TERRITORY:CP_BC"),
        #          :credit_account_uri =>  UU::OS::UESURI.new("ues:TERRITORY:ACCOUNT"),
        #          :is_planned => false,
        #          :credit_business_case_from_opposite_territory => UU::OS::UESURI.new("ues:TERRITORY:ITA_BC"),
        #          :debit_business_case_from_opposite_territory => UU::OS::UESURI.new("ues:TERRITORY:ITA_BC"),
        #          :credit_side_currency => "CZK",
        #          :debit_side_currency => "EUR",
        #          :transaction_currency => "EUR"
        #       )
        def create(control_business_case_uri, payment)
          svc = UU::OS::REST::RemoteClient.new(Payment)
          payment_dto = UU::Finman::FinancialTransaction::PaymentCreate.new(payment)
          # adjust given payment dto to match expected format of financial transaction command
          transaction = transform_create_payment_to_transaction_dto(payment_dto)
          financial_transaction_create = transaction.to_json

          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.post('create', control_business_case_uri, financial_transaction_create)
            if res.nil?
              return res
            else
              return UU::OS::UESURI.new(res)
            end
          end
        end

        # Obtains Payment attributes from service and returns them
        #
        # @param [String, UU::OS::UESURI] payment_uri UESURI of the Payment
        #
        # @return [PaymentAttributes] Payment attributes
        #
        #
        # @example
        #      # Prepare Payment location which we want get attributes.
        #      payment_uri = UU::OS::UESURI.new('ues:TERRITORY:PAYMENT');
        #
        #      # Get Payment
        #      UU::Finman::FinancialTransaction::Payment.get_attributes(payment_uri);
        def get_attributes(payment_uri)
          svc = UU::OS::REST::RemoteClient.new(Payment)
          UU::OS::QoS::QoSHandler.auto_retry do
            attrs_string = svc.get('getAttributes', payment_uri)
            attrs = JSON.parse(attrs_string, :symbolize_names => true)
            attrs = remove_invalid_keys(attrs)
            return UU::Finman::FinancialTransaction::PaymentAttributes.new(attrs)
          end
        end

        # Sets attributes of Payment.
        #
        # @param [String,UU::OS::UESURI] payment_uri UESURI of modified Payment
        # @param [PaymentSetAttributes] payment_set_attributes DTO containing modified attributes of the Payment
        #
        # @return [UU::OS::UESURI] UESURI of modified Payment
        #
        #
        # @example
        #      # Prepare Payment location which we want set attributes.
        #      payment_uri = UU::OS::UESURI.new('ues:TERRITORY:PAYMENT');
        #
        #      # Set Payment
        #      UU::Finman::FinancialTransaction::Payment.set_attributes(payment_uri,
        #        :name => "Payment set");
        def set_attributes(payment_uri, payment_set_attributes)
          svc = UU::OS::REST::RemoteClient.new(Payment)
          payment_dto = UU::Finman::FinancialTransaction::PaymentSetAttributes.new(payment_set_attributes)
          # adjust given payment dto to match expected format of financial transaction command
          transaction = transform_modify_payment_to_transaction_dto(payment_dto)
          financial_transaction_dto = transaction.to_json

          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.post('setAttributes', payment_uri, financial_transaction_dto)
            return UU::OS::UESURI.new(res)
          end
        end


        # Adjust dto of payment to match requirements of financial transaction dto of server side.
        # @private
        def transform_create_payment_to_transaction_dto(payment_dto)
          if payment_dto.is_a?(UU::OS::REST::DTO)
            attributes = payment_dto.to_hash
            # set class of transaction to determine Payment at the server side.
            attributes[:financialTransactionClass] = "PAYMENT"
            attributes[:creditVatDate] = attributes[:vatDate]
            # remove unwanted attributes (in case that e.g. hashmap with attributes from other FinTx was modified and used)
            attributes.delete(:payDate)
            attributes.delete(:metaArtifactUri)
            attributes.delete(:requestApprovalByManager)
            attributes.delete(:waitForAcceptanceByAccountant)
            attributes.delete(:debitVatDate)
            attributes.delete(:vatDate)
            return attributes
          end
          fail "Given object is neither UU::OS::REST::DTO nor its descendant!"
        end
        private :transform_create_payment_to_transaction_dto


        # Hash returned from service contains all attributes from financial transaction.
        # Remove invalid keys to contain only values for this specific class of transaction.
        # @private
        def remove_invalid_keys(hash)
          if hash.is_a?(Hash)
            hash.delete(:complementaryTransferUri)
            hash.delete(:reinvoice_from_category_uri)
            hash.delete(:reinvoice_to_category_uri)
            hash.delete(:reinvoice_from_item)
            hash.delete(:reinvoice_on_item)
            hash.delete(:reinvoicing_type)
            hash.delete(:requestApprovalByManager)
            hash.delete(:waitForAcceptanceByAccountant)
            hash.delete(:closeAfterCreation)
            return hash
          end
          fail "Given object must be Hash!"
        end
        private :remove_invalid_keys


        # Adjust dto of payment to match requirements of financial transaction dto of server side.
        # @private
        def transform_modify_payment_to_transaction_dto(payment_dto)
          if payment_dto.is_a?(UU::OS::REST::DTO)
            attributes = payment_dto.to_hash
            # set class of transaction to determine Payment at the server side.
            attributes[:financialTransactionClass] = "PAYMENT"
            attributes.delete(:payDate)
            attributes.delete(:requestApprovalByManager)
            attributes.delete(:waitForAcceptanceByAccountant)
            return attributes
          end
          fail "Given object is neither UU::OS::REST::DTO nor its descendant!"
        end
        private :transform_modify_payment_to_transaction_dto

      end
    end
  end
end
