require 'uu/os/qos/qos_handler'
require 'uu/os/rest'
require 'uu/os/uesuri'
require 'uu/finman/financialtransaction/payment_copy_attributes'
require 'uu/finman/financialtransaction/payment_copy_set_attributes'

# Module Unicorn Universe
module UU
  # Module Finman
  module Finman
    # Module of finman Financial transaction
    module FinancialTransaction
      #Module of Payment copy
      module PaymentCopy
        extend self

        # Service path
        PATH = 'uu/finman/financialtransaction/FinancialTransaction'

        # Obtains Payment copy attributes from service and returns them
        #
        # @param [String, UU::OS::UESURI] payment_copy_uri UESURI of the Payment
        #
        # @return [PaymentCopyAttributes] Payment copy attributes
        #
        #
        # @example
        #      # Prepare Payment copy location which we want get attributes.
        #      payment_uri = UU::OS::UESURI.new('ues:TERRITORY:PAYMENT_COPY');
        #
        #      # Get Payment
        #      UU::Finman::FinancialTransaction::PaymentCopy.get_attributes(payment_uri);
        def get_attributes(payment_copy_uri)
          svc = UU::OS::REST::RemoteClient.new(PaymentCopy)
          UU::OS::QoS::QoSHandler.auto_retry do
            attrs_string = svc.get('getAttributes', payment_copy_uri)
            attrs = JSON.parse(attrs_string, :symbolize_names => true)
            attrs = remove_invalid_keys(attrs)
            return UU::Finman::FinancialTransaction::PaymentCopyAttributes.new(attrs)
          end
        end

        # Sets attributes of Payment copy.
        #
        # @param [String,UU::OS::UESURI] payment_copy_uri UESURI of modified Payment copy
        # @param [PaymentCopySetAttributes] payment_copy_set_attributes DTO containing modified attributes of the Payment copy
        #
        # @return [UU::OS::UESURI] UESURI of modified Payment copy
        #
        #
        # @example
        #      # Prepare Payment copy location which we want set attributes.
        #      payment_uri = UU::OS::UESURI.new('ues:TERRITORY:PAYMENT_COPY');
        #
        #      # Set Payment copy
        #      UU::Finman::FinancialTransaction::PaymentCopy.set_attributes(payment_copy_uri,
        #        :name => "Payment copy set");
        def set_attributes(payment_copy_uri, payment_copy_set_attributes)
          svc = UU::OS::REST::RemoteClient.new(PaymentCopy)
          payment_copy_dto = UU::Finman::FinancialTransaction::PaymentCopySetAttributes.new(payment_copy_set_attributes)

          # adjust given payment dto to match expected format of financial transaction command
          payment_copy_dto = transform_modify_payment_to_transaction_dto(payment_copy_dto)
          payment_copy_dto = payment_copy_dto.to_json

          UU::OS::QoS::QoSHandler.auto_retry do
            res = svc.post('setAttributes', payment_copy_uri, payment_copy_dto)
            return UU::OS::UESURI.new(res)
          end
        end



        # Hash returned from service contains all attributes from financial transaction.
        # Remove invalid keys to contain only values for specific class of transaction.
        # @private
        def remove_invalid_keys(hash)
          if hash.is_a?(Hash)
            if hash[:originalTransactionClass] == "PAYMENT"
              hash.delete(:complementaryTransferUri)
              hash.delete(:reinvoice_from_category_uri)
              hash.delete(:reinvoice_to_category_uri)
              hash.delete(:reinvoice_from_item)
              hash.delete(:reinvoice_on_item)
              hash.delete(:reinvoicing_type)
              hash.delete(:requestApprovalByManager)
              hash.delete(:waitForAcceptanceByAccountant)
              hash.delete(:closeAfterCreation)
            end

            if hash[:originalTransactionClass] == "FAST_PAYMENT"
              hash.delete(:complementaryTransferUri)
              hash.delete(:reinvoice_from_category_uri)
              hash.delete(:reinvoice_to_category_uri)
              hash.delete(:reinvoice_from_item)
              hash.delete(:reinvoice_on_item)
              hash.delete(:reinvoicing_type)
              hash.delete(:requestApprovalByManager)
              hash.delete(:waitForAcceptanceByAccountant)
              hash.delete(:controlSide)
              hash.delete(:forAcceptanceDate)
              hash.delete(:isPlanned)
            end
            if hash[:originalTransactionClass] == "CREDIT_NOTE"
              hash.delete(:acceptedDate)
              hash.delete(:complementaryTransferUri)
              hash.delete(:forAcceptanceDate)
              hash.delete(:isReadyForAcceptance)
              hash.delete(:reinvoiceFromCategoryUri)
              hash.delete(:reinvoiceToCategoryUri)
              hash.delete(:reinvoiceFromItem)
              hash.delete(:reinvoiceOnItem)
              hash.delete(:reinvoicingType)
              hash.delete(:closeAfterCreation)
            end
            hash.delete(:accountFromOppositeTerritory)
            return hash
          end
          fail "Given object must be Hash!"
        end
        private :remove_invalid_keys

        # Adjust dto of payment to match requirements of financial transaction dto of server side.
        # @private
        def transform_modify_payment_to_transaction_dto(payment_dto)
          if payment_dto.is_a?(UU::OS::REST::DTO)
            attributes = payment_dto.to_hash
            # set class of transaction to determine Payment at the server side.
            attributes[:financialTransactionClass] = "PAYMENT_COPY"
            return attributes
          end
          fail "Given object is neither UU::OS::REST::DTO nor its descendant!"
        end
        private :transform_modify_payment_to_transaction_dto

      end
    end
  end
end
